"""
This python script makes managing the configuration easy
"""

__author__="Parag Mhashilkar <parag@fnal.gov>"
__date__= "May 02 2003"
__version__="1.0"

# $source: BaseConfigManager.py,v$

################################################################################
# Project   : SAM
# Package : jim_config
#
# This work is part of a development project, called JIM, which consists of a
# number of coordinated packages each named jim_xxxx .
#
# Notice of authorship, copyright status,  and terms and conditions, should
# the software eventually become available for use outside Fermilab, can be
# found in the README and LICENCE files in the top level directory of the main
# sam package.
################################################################################

import os, string, sys, time, traceback, re


_JIM_CONFIG_MANAGER_DEBUG_MODE = os.environ.get('JIM_CONFIG_MANAGER_DEBUG_MODE',
                                                'FALSE')
_JIM_CONFIG_MANAGER_LOG = os.environ.get('JIM_CONFIG_MANAGER_LOG')
 
_DEFAULT_MODE = os.environ.get('JIM_CONFIG_MANAGER_DEFAULT_MODE', 'DBFS')

_JIM_XMLDB_CONTEXT = os.environ.get('JIM_XMLDB_CONTEXT')

_JIM_UPS_DB = os.environ.get('JIM_UPS_DB', '.')
 
_HOST_DEPENDENT_PRODUCTS_LIST = ['samgrid', 'jim_advertise',\
                                 'jim_broker', 'jim_broker_client',\
                                 'jim_sandbox', 'jim_info_providers',\
                                 'test_product1']
_PRODUCTS_HOST_NAME = os.environ.get('JIM_CONFIG_MACHINE')
_PRODUCT_CONFIG_DOC_NAME = 'config'
_PRODUCTS_COLLECTION = os.environ.get('JIM_XMLDB_PRODUCTS_COLLECTION',
                                      'products')
 
_SITE_CONFIG_DOC_NAME = 'site_config'
_SITE_COLLECTION = os.environ.get('JIM_XMLDB_SITE_COLLECTION', 'site')
_SITE_PATH = re.compile('//+').sub('/',_JIM_UPS_DB+'/jim_config/site')
 
_GLOBAL_CONFIG_DOC_NAME = 'global_config'
_GLOBAL_COLLECTION = os.environ.get('JIM_XMLDB_GLOBAL_COLLECTION', 'global')
_GLOBAL_PATH = re.compile('//+').sub('/',_JIM_UPS_DB+'/jim_config/global')


################################################################################
################################################################################
# Class: BaseConfigManager
################################################################################
################################################################################
 

class BaseConfigManager:


   class ConfigManagerError(Exception):
      def __init__(self, value):
         self.value = value

      def __str__(self):
         return self.value


   def __init__(self, name):
      """
         Contructor to create a BaseConfigManager

         name - Name of the package
         version - Version of the package.
         qualifiers - Qualifiers for the package

      """
     
      self.name = name
      self.mode = self.getDefaultMode()
      self.path = ''
      self.configDoc = ''
   # __init__


################################################################################
#                              GENERAL FUNCTIONS                               #
################################################################################

   def getDefaultMode(self):
      return _DEFAULT_MODE
   # default_mode

   def getProductsHost(self):
      return _PRODUCTS_HOST_NAME
   # default_mode


   def getHostDependentPrdList(self):
      return _HOST_DEPENDENT_PRODUCTS_LIST
   # hostname_dependent_list  


   def printConfigDetails(self):
      self.printDebug('-----------------------------------------------------------')
      if self.name:
         self.printDebug('My Name               : ' + self.name)
      else:
         self.printDebug('My Name               : ')
      if self.mode:
         self.printDebug('Default Mode          : ' + self.mode)
      else:
         self.printDebug('Default Mode          : ')
      self.printDebug('-----------------------------------------------------------')
   # printConfigDetails


   def printDebug(self, str):
      if (_JIM_CONFIG_MANAGER_LOG) and (not _JIM_CONFIG_MANAGER_LOG == ''):
         try:
            log = open(_JIM_CONFIG_MANAGER_LOG, 'at')
            log.write('JIM_CONFIG:'+time.asctime()+':'+str+'\n')
            log.close()
         except IOError:
            traceback.print_exc()
            print >> sys.stdout, str
      elif (_JIM_CONFIG_MANAGER_DEBUG_MODE == 'TRUE'):
         print >> sys.stdout, str
   # printDebug


   def getDebugMode(self):
      if _JIM_CONFIG_MANAGER_DEBUG_MODE == 'TRUE':
         return 1
      else:
         return 0
   # getDebugMode


################################################################################
#                          PRODUCT RELATED FUNCTIONS                           #
################################################################################

   def getPrdQualifiers(self):
      env_var = string.upper(self.name + '_qualifiers')
      my_qual = os.environ.get(env_var)
      if not my_qual:
         self.printDebug('INFO: Unable to guess ' + env_var)
         self.printDebug('INFO: Using UPS_PROD_QUALIFIERS instead')
         my_qual = os.environ.get('UPS_PROD_QUALIFIERS')
      if not my_qual:
         my_qual = ''
         self.printDebug('INFO: Unable to guess UPS_PROD_QUALIFIERS')
         self.printDebug('INFO: Defaulting to "' + my_qual + '"')
      return my_qual
   # getPrdQualifiers

   def getPrdMachine(self):
      my_machine = self.getProductsHost()
      if not my_machine:
         self.printDebug('INFO: Unable to get a JIM_CONFIG_MACHINE from the env')
         self.printDebug('INFO: Using `hostname` instead')
         import socket
         my_machine =  socket.getfqdn()
      return my_machine
   # getPrdQualifiers

   def getPrdVersion(self):
      env_var = string.upper(self.name + '_version')
      my_ver = os.environ.get(env_var)
      if not my_ver:
         self.printDebug('INFO: Unable to guess ' + env_var)
         self.printDebug('INFO: Using UPS_PROD_VERSION instead')
         my_ver = os.environ.get('UPS_PROD_VERSION')
      if not my_ver:
         my_ver = ''
         self.printDebug('INFO: Unable to guess UPS_PROD_VERSION ')
         self.printDebug('INFO: Defaulting to "' + my_ver + '"')
      return my_ver
   # getPrdVersion


   def getPrdCollection(self):
      context = self.getProductsCollection() + '/' + self.name + '/'\
               + self.version
      if self.qualifiers:
         context = context + '/' + self.qualifiers
      return re.compile('//+').sub('/',context)
   # getPrdCollection


   def getProductsCollection(self):
      context = '/' + _JIM_XMLDB_CONTEXT + '/' + _PRODUCTS_COLLECTION
      return re.compile('//+').sub('/',context)
   # getProductsCollection


   def getPrdConfigDocName(self):
      return _PRODUCT_CONFIG_DOC_NAME
   # configuration_doc_name


   def getPrdPath(self):

      # Check for backward compatibility. When a products version link is
      # created in ${JIM_UPS_DB}/prd_name/Symlinks then the products config
      # must be stored in db else it will be stored in prd area
      link = _JIM_UPS_DB+'/'+self.name+'/Symlinks/'+self.version
      if os.path.islink(link):
         path = _JIM_UPS_DB + '/' + self.name + '/' + self.version + '/'
      else:
         path = self.getPrdReleaseDir() + '/etc'
      return re.compile('//+').sub('/',path)
   # getPrdPath


   def getPrdReleaseDir(self):
      env_var = string.upper(self.name + '_dir')
      my_dir = os.environ.get(env_var)
      if not my_dir:
         self.printDebug('INFO: Unable to guess ' + env_var)
         self.printDebug('INFO: Using UPS_PROD_DIR instead')
         my_dir = os.environ.get('UPS_PROD_DIR')
      if not my_dir:
         my_dir = '.'
         self.printDebug('INFO: Unable to guess UPS_PROD_DIR')
         self.printDebug('INFO: Defaulting to "' + my_dir + '"')
      return my_dir
   # getReleaseDir



################################################################################
#                             SITE RELATED FUNCTIONS                           #
################################################################################

   def getSiteConfigDocName(self):
      return _SITE_CONFIG_DOC_NAME
   # getSiteConfigDocName


   def getSiteContext(self):
      context = '/' + _JIM_XMLDB_CONTEXT + '/' + _SITE_COLLECTION
      return re.compile('//+').sub('/',context)
   # getSiteContext


   def getSitePath(self):
      return _SITE_PATH
   # getSitePath


################################################################################
#                           GLOBAL RELATED FUNCTIONS                           #
################################################################################

   def getGlobalConfigDocName(self):
      return _GLOBAL_CONFIG_DOC_NAME
   # getGlobalConfigDocName


   def getGlobalContext(self):
      context = '/' + _JIM_XMLDB_CONTEXT + '/' + _GLOBAL_COLLECTION
      return re.compile('//+').sub('/',context)
   # getGlobalContext


   def getGlobalPath(self):
      return _GLOBAL_PATH
   # getGlobalPath


################################################################################
#                           CONFIGURATION ACCESS FUNCTIONS                     #
################################################################################


   def getConfigAttr(self, xpath, attr):
      """
         Get a single configuation attribute from the FS
 
         xpath  - Xpath to the attribute
         attr   - Name of the attribute
         returns a single configuration in the form of string
      """
      
      pass
   # getConfigAttr


   def getConfigString(self):
      """
         Get the configuration
 
         return configuration in the form of string
      """
      pass
   # getConfigString


   def getConfigPyXmlObject(self):
      """
         Get the configuration in the form of PyXmlObject

         return configuration in the form of PyXmlObject
      """

      import xml_common_lib
      config_str = self.getConfigString()
      try:
         return xml_common_lib.PyXmlObject.PyXmlObject(config_str)
      except self.ConfigManagerError:
         raise self.ConfigManagerError('Error getting the PyXml object')
   # getConfigPyObject


   def storeConfigFile(self, config_file):
      """
         Store the configuration from the file to default location

         config_file - Name of the configuration file 
         returns 0 if successful, 1 if error
      """

      try:
         f = open(config_file, "rt")
         config_str = f.read(-1)
         f.close()
         return self.storeConfigString(config_str)
      except IOError:
         raise self.ConfigManagerError('File '+config_file+' does not exist')
      except OSError:
         raise self.ConfigManagerError('You do not have permissions to '\
                                       +config_file)
   # storeConfigFile


   def storeConfigString(self, config_str):
      """
         Store the configuration in the form of xml string to default location

         config_str - Configuration in the form of xml string
      """
      pass
   # storeConfigString


# Class BaseConfigManager
