"""
This python script makes managing the configuration in the FS easy
"""

__author__="Parag Mhashilkar <parag@fnal.gov>"
__date__= "May 02 2003"
__version__="1.0"

# $source: BaseFSConfigManager.py,v$

################################################################################
# Project   : SAM
# Package : jim_config
#
# This work is part of a development project, called JIM, which consists of a
# number of coordinated packages each named jim_xxxx .
#
# Notice of authorship, copyright status,  and terms and conditions, should
# the software eventually become available for use outside Fermilab, can be
# found in the README and LICENCE files in the top level directory of the main
# sam package.
################################################################################

import os, string, sys, traceback
import xml_common_lib
from BaseConfigManager import BaseConfigManager


_JIM_CONFIG_FS_CONFIG_FILE = os.environ.get('JIM_CONFIG_FS_CONFIG_FILE')

def getChild(dom,element):
   attribute = None
   value = None
   index = string.find(element,'@') 

   if not  index == -1:
      tmpstr = element[index-1:]
      tmpstr = string.replace(tmpstr,"[@","")
      tmpstr = string.replace(tmpstr,"]","")
      tmpstr = string.replace(tmpstr,"'","")
      tmpstr = string.replace(tmpstr,"\"","")
      tmpstrsplit =  string.split(tmpstr,'=')
      
      attribute = tmpstrsplit[0]
      value = tmpstrsplit[1]
      element   = element[:index-1]


   for f in dom.childNodes:
      if f.nodeName == element:
         if attribute:
            if f.getAttribute(attribute) == value :
	       return f
	 else :
	    return f

   return None
 


def xpath2opath(xpath, attr):
    """
       Convert an xpath string to an 'path' into a PY XML object
       xpath  - xpath to the attibute
       returns the value of the attribute
    """
 
    #Strip off leading and trailing '/'
    if xpath[0] == '/': xpath=xpath[1:]
    if xpath[-1] == '/': xpath=xpath[0:-1]
 
    #XML elements may have '-' but PY objects cannot
    result = string.replace(xpath,'-','_')
 
    # SJ: THIS IS A REALLY BAD ASSUMPTION
    # replacing getConfigAttr with DOM ;-)

    #nested elements correspond to referencing a python attribute
    #We take [0] assuming there is only one element with given name
    result = string.replace(result, '/', '[0].')
 
    return result+'[0].'+attr
#xpath2opath


################################################################################
# Class: BaseFSConfigManager
################################################################################
 

class BaseFSConfigManager(BaseConfigManager):
   """This utility class manages the configuration access to Db"""

   def __init__(self, name):
      """
         Contructor to create a BaseFSConfigManager

         name    - Name of the package
      """

      BaseConfigManager.__init__(self, name)
   # __init__


   def printConfigDetails(self):
      self.printDebug('-----------------------------------------------------------')
      self.printDebug('My Name               : ' + self.name)
      self.printDebug('My config path        : ' + self.path)
      self.printDebug('Name of configuration : ' + self.configDoc)
      self.printDebug('-----------------------------------------------------------')
   # printConfigDetails


   def getConfigAttr(self, xpath, attr):
      """
         Get a single configuation attribute from the FS

         xpath  - Xpath to the attribute
         attr   - Name of the attribute
         returns a single configuration attribute value in the form of string
      """

      file_path = self.path + '/' + self.configDoc
      if not os.path.exists(file_path):
          self.printDebug("No  file as %s" % file_path) 
	  file_path = self.path + '/' + "config"
	  self.printDebug("Using %s instead" % file_path)
	  
      try :
          from xml.dom.minidom import parse
	  dom = parse(file_path)
          
	  #Strip off leading and trailing '/'
          if xpath[0] == '/': xpath=xpath[1:]
          if xpath[-1] == '/': xpath=xpath[0:-1]
	  for element in string.split(xpath,'/'):
              dom = getChild(dom,element)
              if dom == None:
                 print "element %s not found" % element
	         raise Exception("element %s not found" % element)
	  
	  return dom.getAttribute(attr)
      except Exception,ex:
          raise ex 
      
   # getConfigAttr


   def getConfigString(self):
      """
         Get the configuration from FS
 
         returns configuration in the form of string
      """

      file_path = self.path + '/' + self.configDoc
      if not os.path.exists(file_path):
          self.printDebug("No  file as %s" % file_path) 
	  file_path = self.path + '/' + "config"
	  self.printDebug("Using %s instead" % file_path)
	  
      try:
         self.printDebug('Retrieving the configuration file ' + file_path)
         f = open(file_path,"rt")
         config = f.read(-1)
         f.close()
         return config
      except:
         if self.getDebugMode(): traceback.print_exc()
         raise self.ConfigManagerError('error retrieving the config from FS')
   # getConfigString


   def storeConfigString(self, config_str):
      """
         Store the configuration in the form of xml string to the FS

         config_str - Configuration in the form of xml string
         returns 0 if successful, 1 if error occurs
      """

      # Create the directiry if it does not exist
      # Specially useful for creating global and site directories
      if not os.path.exists(self.path):
         os.system('mkdir -p ' + self.path)

      file_path = self.path + '/' + self.configDoc
      self.printDebug('Checking for write access to '+self.path+' ...\n')
      if os.access(self.path, os.W_OK):
         try:
            f = open(file_path, "wt")
            f.write(config_str)
            self.printDebug('Your configuration was successfully stored to '\
                        + file_path + ' ...\n')
            return 0
         except:
            if self.getDebugMode(): traceback.print_exc()
            raise self.ConfigManagerError('error storing config to FS')
      else:
         print >> sys.stderr, 'You do not have permissions to write to '\
                              + self.path + '\n'
         raise self.ConfigManagerError('insufficient privilages to store '\
                                       + 'the config in FS')
   # storeConfigString

################################################################################
#                           GLOBAL RELATED FUNCTIONS                           #
################################################################################

   def getGlobalConfigDocName(self):
      if _JIM_CONFIG_FS_CONFIG_FILE:
        import os.path
        name = os.path.basename(_JIM_CONFIG_FS_CONFIG_FILE)
      else:
        name = BaseConfigManager.getGlobalConfigDocName(self)
      return name
   # getGlobalConfigDocName

   def getGlobalPath(self):
      if _JIM_CONFIG_FS_CONFIG_FILE:
        import os.path
        path = os.path.dirname(_JIM_CONFIG_FS_CONFIG_FILE)
      else:
        path = BaseConfigManager.getGlobalPath(self)  
      return path
   # getGlobalPath

################################################################################
#                             SITE RELATED FUNCTIONS                           #
################################################################################

   def getSiteConfigDocName(self):
      if _JIM_CONFIG_FS_CONFIG_FILE:
        import os.path
        name = os.path.basename(_JIM_CONFIG_FS_CONFIG_FILE)
      else:
        name = BaseConfigManager.getSiteConfigDocName(self)
      return name
   # getSiteConfigDocName


   def getSitePath(self):
      if _JIM_CONFIG_FS_CONFIG_FILE:
        import os.path
        path = os.path.dirname(_JIM_CONFIG_FS_CONFIG_FILE)
      else:
        path = BaseConfigManager.getSitePath(self)  
      return path
   # getSitePath


################################################################################
#                          PRODUCT RELATED FUNCTIONS                           #
################################################################################

   def getPrdPath(self):

      # Override defaults if path to a specific config file is provided
      if _JIM_CONFIG_FS_CONFIG_FILE:
        import os.path
        prdPath = os.path.dirname(_JIM_CONFIG_FS_CONFIG_FILE)
      else:
        prdPath = BaseConfigManager.getPrdPath(self)
      return prdPath
   # getPrdPath


   def getPrdConfigDocName(self):
      if _JIM_CONFIG_FS_CONFIG_FILE:
        import os.path
        name = os.path.basename(_JIM_CONFIG_FS_CONFIG_FILE)
      else:
        name = BaseConfigManager.getPrdConfigDocName(self)
        for prd in self.getHostDependentPrdList():
           if prd == self.name:
              name = name + '.' + self.machine
              break
      return name
   # configuration_doc_name

# Class BaseFSConfigManager
