"""
This python library makes managing the configuration easy
"""

__author__="Parag Mhashilkar <parag@fnal.gov>"
__date__= "May 15 2003"
__version__="1.0"

# $source: JimConfigManager.py,v$

################################################################################
# Project   : SAM
# Package : jim_config
#
# This work is part of a development project, called JIM, which consists of a
# number of coordinated packages each named jim_xxxx .
#
# Notice of authorship, copyright status,  and terms and conditions, should
# the software eventually become available for use outside Fermilab, can be
# found in the README and LICENCE files in the top level directory of the main
# sam package.
################################################################################

import os, string, sys, traceback
import BaseConfigManager
import GlobalConfigManager
import SiteConfigManager
import ProductConfigManager


################################################################################
# Class: JimConfigManager
################################################################################
 

class JimConfigManager:

   def __init__(self, name=None, version=None, qualifiers=None):
      """
         Contructor to create a JimConfigManager

         name - Name of the package.
                Can be None for accessing site or global configuration
         version - Version of the package.
         qualifiers - Qualifiers for the package

         If version and qualifiers are None, they are assigned to
         UPS_PROD_VERSION and UPS_PROD_QUALIFIERS environment variables.
      """
     
      self.name = name
      self.version = version
      self.qualifiers = qualifiers

      conf_man = BaseConfigManager.BaseConfigManager(name)
      self.debug = conf_man.getDebugMode()
      self.mode = conf_man.getDefaultMode()

      #print '---------------------------------------------------------------'
      #print 'Product Name          : ', self.name
      #print 'Product Version       : ', self.version
      #print 'Product Qualifiers    : ', self.qualifiers
      #print '---------------------------------------------------------------'
   # __init__


   def setMode(self, mode):
      """
         Override the default mode 'DBFS'.In this mode it access first database
         and in case of errors retrieving configuation it accesses file system
         To use only file system use 'FS'.

         mode - Mode to access the information
      """

      self.mode = mode
   # setMode


   def getDebugMode(self):
      return self.debug
   # getDebugMode

   
   def printDBLookupErrorMessage(self):
     print >> sys.stderr,'\nWARNING: DB lookup for %s failed; using FS instead' % (self.name)
        
   def printFSLookupErrorMessage(self):
     print >> sys.stderr,'\nERROR: Could not find the config of %s in FS' % (self.name)

   def printDBStorageErrorMessage(self):
     print >> sys.stderr, '\nWARNING: DB storage of %s failed' % (self.name)

################################################################################
#                             PRODUCT RELATED FUNCTIONS                        #
################################################################################


   def getConfigAttr(self, xpath, attr):
      """
         Get a single configuation attribute

         xpath - Xpath to the attribute
         attr - Name of the attribute
         return a single configuration in the form of string
      """

      import ProductConfigManager

      mode = self.mode
      if mode == 'DBFS':
         try:
            jcm_db = ProductConfigManager.ProductDBConfigManager(
                        self.name, self.version, self.qualifiers)
            value = jcm_db.getConfigAttr(xpath, attr)
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            if self.debug: traceback.print_exc()
            self.printDBLookupErrorMessage()
            mode = 'FS'

      if mode == 'FS':
         try:
            jcm_fs = ProductConfigManager.ProductFSConfigManager(
                        self.name, self.version, self.qualifiers)
            value = jcm_fs.getConfigAttr(xpath, attr)
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            self.printFSLookupErrorMessage()
            print >> sys.stderr,'     : Have you tailored this package?' 
            print >> sys.stderr,'     : Is the attribute path correct?' 
            raise

      return value
   # getConfigAttr


   def getConfigString(self):
      """
         Get the configuration
 
         return configuration in the form of string
      """

      import ProductConfigManager

      mode = self.mode
      if mode == 'DBFS':
         try:
            jcm_db = ProductConfigManager.ProductDBConfigManager(
                        self.name, self.version, self.qualifiers)
            value = jcm_db.getConfigString()
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            if self.debug: traceback.print_exc()
            self.printDBLookupErrorMessage()
            mode = 'FS'

      if mode == 'FS':
         try:
            jcm_fs = ProductConfigManager.ProductFSConfigManager(
                        self.name, self.version, self.qualifiers)
            value = jcm_fs.getConfigString()
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            self.printFSLookupErrorMessage()
            print >> sys.stderr,'\nERROR: Have you tailored this package?' 
            raise

      return value
   # getConfigString


   def getConfigPyXmlObject(self):
      """
         Get the configuration in the form of PyXmlObject

         return configuration in the form of PyXmlObject
      """

      import ProductConfigManager

      mode = self.mode
      if mode == 'DBFS':
         try:
            jcm_db = ProductConfigManager.ProductDBConfigManager(
                        self.name, self.version, self.qualifiers)
            value = jcm_db.getConfigPyXmlObject()
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            if self.debug: traceback.print_exc()
            self.printDBLookupErrorMessage()
            mode = 'FS'

      if mode == 'FS':
         try:
            jcm_fs = ProductConfigManager.ProductFSConfigManager(
                        self.name, self.version, self.qualifiers)
            value = jcm_fs.getConfigPyXmlObject()
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            self.printFSLookupErrorMessage()
            print >> sys.stderr,'\nERROR: Have you tailored this package?' 
            raise

      return value
   # getConfigPyXmlObject


   def storeConfigFile(self, config_file):
      """
         Store the configuration from the file

         config_file - Name of the configuration file 
      """

      # Configuration is always stored into the FS first
      mode = 'FS'
      if mode == 'FS':
         try:
            jcm_fs = ProductConfigManager.ProductFSConfigManager(
                        self.name, self.version, self.qualifiers)
            jcm_fs.storeConfigFile(config_file)
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            raise

      mode = self.mode
      if mode == 'DBFS':
         try:
            jcm_db = ProductConfigManager.ProductDBConfigManager(
                        self.name, self.version, self.qualifiers)
            jcm_db.storeConfigFile(config_file)
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            if self.debug: traceback.print_exc()
            self.printDBStorageErrorMessage()

   # storeConfigFile


   def storeConfigString(self, config_str):
      """
         Store the configuration in the form of xml string

         config_str - Configuration in the form of xml string
      """

      # Configuration is always stored into the FS first
      mode = 'FS'
      if mode == 'FS':
         try:
            jcm_fs = ProductConfigManager.ProductFSConfigManager(
                        self.name, self.version, self.qualifiers)
            jcm_fs.storeConfigString(config_str)
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            raise

      mode = self.mode
      if mode == 'DBFS':
         try:
            jcm_db = ProductConfigManager.ProductDBConfigManager(
                        self.name, self.version, self.qualifiers)
            jcm_db.storeConfigString(config_str)
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            if self.debug: traceback.print_exc()
            self.printDBStorageErrorMessage()

   # storeConfigString


################################################################################
#                               SITE RELATED FUNCTIONS                         #
################################################################################


   def getSiteConfigAttr(self, xpath, attr):
      """
         Get a global configuation attribute
 
         xpath - Xpath to the attribute
         attr - Name of the attribute
         return a site configuration in the form of string
      """
 
      mode = self.mode
      if mode == 'DBFS':
         try:
            jcm_db = SiteConfigManager.SiteDBConfigManager(self.name)
            value = jcm_db.getConfigAttr(xpath, attr)
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            if self.debug: traceback.print_exc()
            self.printDBLookupErrorMessage()
            mode = 'FS'

      if mode == 'FS':
         try:
            jcm_fs = SiteConfigManager.SiteFSConfigManager(self.name)
            value = jcm_fs.getConfigAttr(xpath, attr)
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            self.printFSLookupErrorMessage()
            print >> sys.stderr,'     : Have you configured the site?' 
            print >> sys.stderr,'     : Is the attribute path correct?' 
            raise
      return value
   # getSiteConfigAttr


   def getSiteConfigString(self):
      """
         Get the site configuration in the form of string

         return site configuration in the form of string
      """

      mode = self.mode
      if mode == 'DBFS':
         try:
            jcm_db = SiteConfigManager.SiteDBConfigManager(self.name)
            value = jcm_db.getConfigString()
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            if self.debug: traceback.print_exc()
            self.printDBLookupErrorMessage()
            mode = 'FS'

      if mode == 'FS':
         try:
            jcm_fs = SiteConfigManager.SiteFSConfigManager(self.name)
            value = jcm_fs.getConfigString()
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            self.printFSLookupErrorMessage()
            print >> sys.stderr,'\nERROR: Have you configured the site?' 
            raise

      return value
   # getSiteConfigString


   def getSiteConfigPyXmlObject(self):
      """
         Get the site configuration in the form of PyXmlObject

         return site configuration in the form of PyXmlObject
      """

      mode = self.mode
      if mode == 'DBFS':
         try:
            jcm_db = SiteConfigManager.SiteDBConfigManager(self.name)
            value = jcm_db.getConfigPyXmlObject()
         except Exception:
            if self.debug: traceback.print_exc()
            self.printDBLookupErrorMessage()
            mode = 'FS'

      if mode == 'FS':
         try:
            jcm_fs = SiteConfigManager.SiteFSConfigManager(self.name)
            value = jcm_fs.getConfigPyXmlObject()
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            self.printFSLookupErrorMessage()
            print >> sys.stderr,'\nERROR: Have you configured the site?' 
            raise

      return value
   # getSiteConfigPyXmlObject


   def storeSiteConfigFile(self, config_file):
      """
         Store the site configuration from the file
 
         config_file - Name of the configuration file
      """
 
      mode = 'FS'
      if mode == 'FS':
         try:
            jcm_fs = SiteConfigManager.SiteFSConfigManager(self.name)
            value = jcm_fs.storeConfigFile(config_file)
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            raise
      mode = self.mode
      if mode == 'DBFS':
         try:
            jcm_db = SiteConfigManager.SiteDBConfigManager(self.name)
            value = jcm_db.storeConfigFile(config_file)
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            if self.debug: traceback.print_exc()
            self.printDBStorageErrorMessage()
      return value
   # storeSiteConfigFile


   def storeSiteConfigString(self, config_str):
      """
         Store the site configuration in the form of xml string 

         config_str - Configuration in the form of xml string
      """

      mode = 'FS'
      if mode == 'FS':
         try:
            jcm_fs = SiteConfigManager.SiteFSConfigManager(self.name, self.version,
                                            self.qualifiers)
            value = jcm_fs.storeConfigString(config_str)
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            raise

      mode = self.mode
      if mode == 'DBFS':
         try:
            jcm_db = SiteConfigManager.SiteDBConfigManager(self.name, self.version,
                                            self.qualifiers)
            value = jcm_db.storeConfigString(config_str)
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            if self.debug: traceback.print_exc()
            self.printDBStorageErrorMessage()
      return value
   # storeSiteConfigString


   def getSiteConfiguration(self):
      """
         Get the site configuration in the form of string
 
         return site configuration in the form of string
      """

      return self.getSiteConfigString()
   # getSiteConfiguration


################################################################################
#                             GLOBAL RELATED FUNCTIONS                         #
################################################################################


   def getGlobalConfigAttr(self, xpath, attr):
      """
         Get a global configuation attribute
 
         xpath - Xpath to the attribute
         attr - Name of the attribute
         return a global configuration in the form of string
      """
 
      mode = self.mode
      if mode == 'DBFS':
         try:
            jcm_db = GlobalConfigManager.GlobalDBConfigManager(self.name)
            value = jcm_db.getConfigAttr(xpath, attr)
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            if self.debug: traceback.print_exc()
            self.printDBLookupErrorMessage()
            mode = 'FS'

      if mode == 'FS':
         try:
            jcm_fs = GlobalConfigManager.GlobalFSConfigManager(self.name)
            value = jcm_fs.getConfigAttr(xpath, attr)
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            self.printFSLookupErrorMessage()
            print >> sys.stderr,'     : Have you stored the global constants?'
            print >> sys.stderr,'     : Is the attribute path correct?' 
            raise
      return value
   # getGlobalConfigAttr


   def getGlobalConfigString(self):
      """
         Get the global configuration in the form of string

         return global configuration in the form of string
      """

      mode = self.mode
      if mode == 'DBFS':
         try:
            jcm_db = GlobalConfigManager.GlobalDBConfigManager(self.name)
            value = jcm_db.getConfigString()
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            if self.debug: traceback.print_exc()
            print >> sys.stderr, '\nWARNING: DB lookup failed using FS instead\n'
            mode = 'FS'

      if mode == 'FS':
         try:
            jcm_fs = GlobalConfigManager.GlobalFSConfigManager(self.name)
            value = jcm_fs.getConfigString()
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            self.printFSLookupErrorMessage()
            print >> sys.stderr,'\nERROR: Have you stored the global constants?'
            raise
      return value
   # getGlobalConfigString


   def getGlobalConfigPyXmlObject(self):
      """
         Get the global configuration in the form of PyXmlObject

         return global configuration in the form of PyXmlObject
      """

      mode = self.mode
      if mode == 'DBFS':
         try:
            jcm_db = GlobalConfigManager.GlobalDBConfigManager(self.name)
            value = jcm_db.getConfigPyXmlObject()
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            if self.debug: traceback.print_exc()
            self.printDBLookupErrorMessage()
            mode = 'FS'

      if mode == 'FS':
         try:
            jcm_fs = GlobalConfigManager.GlobalFSConfigManager(self.name)
            value = jcm_fs.getConfigPyXmlObject()
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            self.printFSLookupErrorMessage()
            print >> sys.stderr,'\nERROR: Have you stored the global constants?'
            raise
      return value
   # getGlobalConfigPyXmlObject


   def storeGlobalConfigFile(self, config_file):
      """
         Store the global configuration from the file
 
         config_file - Name of the configuration file
      """
 
      mode = 'FS'
      if mode == 'FS':
         try:
            jcm_fs = GlobalConfigManager.GlobalFSConfigManager(self.name)
            value = jcm_fs.storeConfigFile(config_file)
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            raise

      mode = self.mode
      if mode == 'DBFS':
         try:
            jcm_db = GlobalConfigManager.GlobalDBConfigManager(self.name)
            value = jcm_db.storeConfigFile(config_file)
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            if self.debug: traceback.print_exc()
            self.printDBStorageErrorMessage()
      return value
   # storeGlobalConfigFile


   def storeGlobalConfigString(self, config_str):
      """
         Store the global configuration in the form of xml string

         config_str - Configuration in the form of xml string
      """

      mode = 'FS'
      if mode == 'FS':
         try:
            jcm_fs = GlobalConfigManager.GlobalFSConfigManager(self.name)
            value = jcm_fs.storeConfigString(config_str)
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            raise

      mode = self.mode
      if mode == 'DBFS':
         try:
            jcm_db = GlobalConfigManager.GlobalDBConfigManager(self.name)
            value = jcm_db.storeConfigString(config_str)
         except BaseConfigManager.BaseConfigManager.ConfigManagerError:
            if self.debug: traceback.print_exc()
            self.printDBStorageErrorMessage()

   # storeGlobalConfigString


# Class JimConfigManager
