#!/usr/bin/env sampy
################################################################################
# Project   : SAM
# Package : sam_cp
# $Id: sam_cp.py,v
# 1.29 2001/06/04 21:39:09 lauri Exp $
#
# This work is part of a development project, called SAM, which consists of a
# number of coordinated packages each named sam_xxxx .
#
# Notice of authorship, copyright status,  and terms and conditions, should
# the software eventually become available for use outside Fermilab, can be
# found in the README and LICENCE files in the top level directory of the main
# sam package.
################################################################################

import os
import sys
import string
import time
import re 
import random
import stat

from SamCpFileParser import SamCpFileParser
import SamCpExceptions 

from SamCpWrapper import SRC_, DEST_, DEBUG_, VERBOSE_
import SamCpOutputHandler
import SamCpEnstoreFileInfoHandler


from SamUtility.CommonUtility import ExecSubprocess
from SamUtility.SamUserIdentifier import getUnixUsername

##############################################################################
# Constants.



################################################################################
class SamCpProtocolBaseClass:
    """
    Base class (aka interface) from which all other protocols are derived.
    """
    def __init__(self, argDict={}):
        self.parseArgs(argDict)

    def parseArgs(self, argDict):
        self.debug = argDict.get(DEBUG_)
        self.verbose = argDict.get(VERBOSE_)

        srcFile = argDict.get(SRC_)
        if not isinstance(srcFile, SamCpFileParser):
            srcFile = SamCpFileParser(srcFile)
        self.src = srcFile

        destFile = argDict.get(DEST_)
        if not isinstance(destFile, SamCpFileParser):
            destFile = SamCpFileParser(destFile)
        self.dest = destFile

    #########################################################################
    def log(self, flag, msgPrefix, *args):
        if flag:
            for arg in args:
                print("%s%s" % (msgPrefix, str(arg)))
    def dbgprint(self, args):
        now = time.strftime( '%H:%M:%S', time.localtime(time.time()) )
        return self.log(self.debug, "<%s> " % now, args)
    def vsay(self, args): return self.log(self.verbose, '', args)
    #########################################################################
    def copy(self):
        try:
            self.authenticate()
            self.prepareFileTransfer()
            self.performFileTransfer()
            self.postProcessFileTransfer()
            self.setFileTransferStatus()
            returnStatus = self.getFileTransferStatus()
        except:
            raise
        return returnStatus
    ###########################################################################
    # DERIVED CLASSES MUST IMPLEMENT THESE METHODS IF THEY ARE USED!
    def authenticate(self):            raise SamCpExceptions.SamCpMethodNotImplementedException()
    def prepareFileTransfer(self):     raise SamCpExceptions.SamCpMethodNotImplementedException()
    def performFileTransfer(self):     raise SamCpExceptions.SamCpMethodNotImplementedException()
    def postProcessFileTransfer(self): raise SamCpExceptions.SamCpMethodNotImplementedException()
    def setFileTransferStatus(self):   raise SamCpExceptions.SamCpMethodNotImplementedException()
    def getFileTransferStatus(self):   raise SamCpExceptions.SamCpMethodNotImplementedException()
    ###########################################################################    
################################################################################

class SamCpExecvpShellCommand(SamCpProtocolBaseClass):
    """
    Basic implementation where we execvp a shell command, switching
    context to that command and never return.
    """
    def __init__(self, args):
        SamCpProtocolBaseClass.__init__(self, args)

        self.className = self.__class__.__name__
        self.command = "${SAM_CP_DIR}/test/echoInputParams.sh"

    ###########################################################################    
    def authenticate(self):
        """
        No authentication performed in this class.
        """
        pass

    ###########################################################################    
    def prepareFileTransfer(self):
        """
        Convert the file objects to strings that will be used
        in constructing the command
        """
        self.srcFile = self.src.parsedFileName()
        self.destFile = self.dest.parsedFileName()

        self.dbgprint("prepareFileTransfer: srcFile = %s" % self.srcFile)
        self.dbgprint("prepareFileTransfer: destFile = %s" % self.destFile)

        self.args = []
        self.dbgprint("prepareFileTransfer: args = %s" % str(self.args))
    ###########################################################################    
    def performFileTransfer(self):
        # we need to expand any environmental variables
        # in order for os.execvp to understand them:
        cmd  = os.path.expandvars(self.command)

        # the way os.execvp operates:
        # os.execvp(cmd, args)
        # where
        #   cmd is the path to the command
        #   args[0] is the C-style argv[0], i.e., the name
        #    of the command internally, NOT an actual 'argument' as seen by the user
        #   args[1:] are the command line args.
        args = ["%s" % os.path.basename(cmd)]
        for arg in self.args:
            args.append(os.path.expandvars(arg))

        src  = os.path.expandvars("%s" % self.srcFile)
        dest = os.path.expandvars("%s" % self.destFile)
            
        args.append(src)
        args.append(dest)

        if ( self.getDebugArgs() ):
            args.append(self.getDebugArgs())

        if ( self.getVerboseArgs() ):
            args.append(self.getVerboseArgs())
        
        try:
            self.dbgprint("%s exec> %s %s" % (self.className, cmd, string.join(args[1:])))
            self.vsay("Now executing command: %s" % string.join(args))
            # if the execvp succeeds, then the process context is switched
            #  to that process, and the caller will receive the return
            #  status of THAT process.
            os.execvp(cmd, args)
        except:
            # but if we failed, for any reason, we gotta pass it on up.
            raise SamCpExceptions.SamCpFailure("%s %s" % (cmd, string.join(args[1:])), sys.exc_info()[1])

    ###########################################################################    
    # placeholder to be filled in by subclasses that know
    # how to turn on debugging
    def getDebugArgs(self):
        return None

    ###########################################################################    
    # placeholder to be filled in by subclasses that know
    # how to turn on verbosity
    def getVerboseArgs(self):
        return None
################################################################################
class SamCp(SamCpExecvpShellCommand):
    """
    For backwards compatibility with existing SAM_CP_CONFIG_FILE settings
    """
    pass
################################################################################
#
# Rcp: whatever is in the path.
#
class SamRcp(SamCpExecvpShellCommand):
    def __init__(self, args={}):
        SamCpExecvpShellCommand.__init__(self, args)
        self.command = "rcp"

################################################################################
#
# Scp: whatever is in the path.
#
class SamScp(SamCpExecvpShellCommand):
    def __init__(self, args={}):
        SamCpExecvpShellCommand.__init__(self, args)
        self.command = "scp"

################################################################################
#
# Local cp:
#
class SamLocalCp(SamCpExecvpShellCommand):
    def __init__(self, args={}):
        SamCpExecvpShellCommand.__init__(self, args)
        self.command = "cp"

    def prepareFileTransfer(self):
        self.srcFile = self.src.path()
        self.destFile = self.dest.path()
        self.args = []

################################################################################
#
# Sam_bbftp:
#
class SamBbftp(SamCpExecvpShellCommand):
    def __init__(self, args={}):
        SamCpExecvpShellCommand.__init__(self, args)
        self.command = "${SAM_BBFTP_DIR}/bin/bbrcp"

################################################################################
#
# Sam_gridftp:
#
class SamGridftp(SamCpExecvpShellCommand):
    def __init__(self, args={}):
        SamCpExecvpShellCommand.__init__(self, args)
        self.command = "${SAM_GRIDFTP_DIR}/bin/gridftp"
            
################################################################################
#
# jim_gridftp:
#
class JimGridftp(SamCpExecvpShellCommand):
    def __init__(self, args={}):
        SamCpExecvpShellCommand.__init__(self, args)
        self.command = "${JIM_GRIDFTP_DIR}/bin/jim_gridftp"
            

################################################################################
#
# sam_fcp:
#
class SamFcp(SamCpExecvpShellCommand):
    def __init__(self, args={}):
        SamCpExecvpShellCommand.__init__(self, args)
        self.command = "${SAM_FCP_DIR}/bin/sam_fcp.py"


################################################################################
#
# Sam_kerberos_rcp:
#
class SamKerberosRcp(SamCpExecvpShellCommand):
    def __init__(self, args={}):
        SamCpExecvpShellCommand.__init__(self, args)
        self.command = "${SAM_KERBEROS_RCP_DIR}/bin/sam_krcp.sh"

################################################################################
#
# Enstore:
#
class SamEnstore(SamCpExecvpShellCommand):
    def __init__(self, args={}):
        SamCpExecvpShellCommand.__init__(self, args)

        # CDF dcache stuff: READING from their robot
        if ( string.find(self.src.parsedFileName(), "/pnfs/cdfen/")  != -1 ):
            self.command = "${SAM_DCACHE_CP_DIR}/bin/sam_dcache_cp"

        # D0 standard encp wrappers; also used for CDF WRITE to robot
        else:
            self.command = "${SAM_ENCP_DIR}/bin/sam_encp"

# specifically for D0:
class SamEncp(SamEnstore):
    def __init__(self, args={}):
        SamEnstore.__init__(self, args)
        self.command = "${SAM_ENCP_DIR}/bin/sam_encp"

class SamDccp(SamEnstore):
    def __init__(self, args={}):
        SamEnstore.__init__(self, args)
        self.command = "${SAM_DCCP_DIR}/bin/sam_dccp.sh"

# specifically for CDF:
class SamDcache(SamEnstore):
    def __init__(self, args={}):
        SamEnstore.__init__(self, args)
        self.command = "${SAM_DCACHE_CP_DIR}/bin/sam_dcache_cp"

class SamDcacheGridftp(SamEnstore):
    def __init__(self, args={}):
        SamCpExecvpShellCommand.__init__(self, args)
        self.command = "${SAM_DCACHE_CP_DIR}/bin/sam_dcache_gridftp"

class SamRfioBase(SamCpExecvpShellCommand):
    def __init__(self, args={}):
        SamCpExecvpShellCommand.__init__(self, args)

    def prepareFileTransfer(self):
        self.srcFile = self.hpssConvert(self.src.parsedFileName())
        self.destFile = self.hpssConvert(self.dest.parsedFileName())
        self.args = []
        
    def hpssConvert(self,path):
        import string
        if ( string.find(path,"rfio://") != -1 ):
            return SamCpFileParser(path).path()
        return path

class SamRfio(SamRfioBase):
    def __init__(self, args={}):
        SamCpExecvpShellCommand.__init__(self, args)
        self.command = "rfcp"
     
class SamBbftpRfio(SamRfioBase):
    def __init__(self, args={}):
        SamCpExecvpShellCommand.__init__(self, args)
        self.command = "${SAM_BBFTP_DIR}/bin/bbrcp"

################################################################################
# for testing TSM robot at d0karlsruhe
class SamGridKaCp(SamCpExecvpShellCommand):
    def __init__(self, args={}):
        SamCpExecvpShellCommand.__init__(self, args)
        self.command = "GridKaCp"
        
################################################################################
# for testing exception handling
class SamDummyCp(SamCpProtocolBaseClass): pass
################################################################################


################################################################################

class SamCpExecSubprocess(SamCpProtocolBaseClass):
    """
    Basic implementation where we execute subprocess.
    """
    def __init__(self, args):
        SamCpProtocolBaseClass.__init__(self, args)

        self.className = self.__class__.__name__
        self.command = "${SAM_CP_DIR}/test/echoInputParams.sh"
        self.exitStatus = None
        self.args = args
        self.fullCommand = None
        self.outputHandler = SamCpOutputHandler.SamCpOutputHandler()
        

    #######################################################################
    # Hack to get non-zero exit status.
    def copy(self):
        exitStatus = SamCpProtocolBaseClass.copy(self)
        cmd = "sh"
        args = ["%s" % os.path.basename(cmd), "-c", "exit %s" % (exitStatus)]
        self.dbgprint("%s exiting with status %s" % (self.className, 
                                                     exitStatus))
        os.execvp(cmd, args)

    #######################################################################
    def authenticate(self):
        """
        No authentication performed in this class.
        """
        pass

    ###########################################################################    
    def prepareFileTransfer(self):
        """
        Convert the file objects to strings that will be used
        in constructing the command
        """
        self.srcFile = self.src.parsedFileName()
        if self.src.isVirtualDomain():
          self.srcFile = self.src.path()

        self.destFile = self.dest.parsedFileName()
        if self.dest.isVirtualDomain():
          self.destFile = self.dest.path()

        self.dbgprint("prepareFileTransfer: srcFile = %s" % self.srcFile)
        self.dbgprint("prepareFileTransfer: destFile = %s" % self.destFile)

    ###########################################################################    
    def performFileTransfer(self):
        # Check if the full command has been set.
        # If not, create the command.
        if ( not self.fullCommand ):
            # we need to expand any environmental variables
            cmd  = os.path.expandvars(self.command)


            args = ["%s" % os.path.basename(cmd)]

            src  = os.path.expandvars("%s" % self.srcFile)
            dest = os.path.expandvars("%s" % self.destFile)
            
            args.append(src)
            args.append(dest)
        
            if ( self.getDebugArgs() ):
                args.append(self.getDebugArgs())

            if ( self.getVerboseArgs() ):
                args.append(self.getVerboseArgs())
            self.fullCommand = "%s %s" % (cmd, string.join(args[1:]))
        
        exitStatus = 0
        stdErr = ""
        stdOut = ""
        try:
            fullCmd = self.fullCommand
            self.dbgprint("%s exec> %s" % (self.className, fullCmd))
            (stdOut, stdErr, exitStatus) = ExecSubprocess(fullCmd)
            self.dbgprint("%s perform file transfer: exitStatus: %s" % (
                self.className, exitStatus))
            self.dbgprint("%s perform file transfer: stdOut: %s" % (
                self.className, stdOut))
            self.dbgprint("%s perform file transfer: stdErr: %s" % (
                self.className, stdErr))
            outputDict = {
              SamCpOutputHandler.FILE_TRANSFER_EXIT_STATUS_KEY_ : exitStatus,
              SamCpOutputHandler.FILE_TRANSFER_STD_ERR_KEY_ : stdErr,
              SamCpOutputHandler.FILE_TRANSFER_STD_OUT_KEY_ : stdOut,
              }
            self.outputHandler.add(outputDict)
              
        except:
            # but if we failed, for any reason, we gotta pass it on up.
            msg = "File transfer failed: %s" % (sys.exc_info()[1])
            raise SamCpExceptions.SamCpFailure(msg)

        if ( exitStatus ):
            msg = "File transfer failed: %s " % (stdErr)
            raise SamCpExceptions.SamCpFileTransferFailure(msg)

        if not self.exitStatus:
            self.exitStatus = exitStatus
        else:
            self.exitStatus = self.exitStatus | exitStatus

    ###########################################################################    
    # placeholder to be filled in by subclasses that know
    # how to post-process a file
    def postProcessFileTransfer(self):
        self.dbgprint("%s post process file transfer" % (self.className))
        self.outputHandler.writeOutput()
    
    ###########################################################################    
    # placeholder to be filled in by subclasses that know
    # what to do with a file transfer status
    def setFileTransferStatus(self):
        pass
    
    ###########################################################################    
    # placeholder to be filled in by subclasses that know
    # what to do with a file transfer status
    def getFileTransferStatus(self):
        self.dbgprint("%s get file transfer status: exitStatus %s" % (
          self.className, self.exitStatus))
        return self.exitStatus
    
    ###########################################################################    
    # placeholder to be filled in by subclasses that know
    # how to turn on debugging
    def getDebugArgs(self):
        return None

    ###########################################################################    
    # placeholder to be filled in by subclasses that know
    # how to turn on verbosity
    def getVerboseArgs(self):
        return None


############################################################################# 
# New enstore class.

class SamCpEnstore(SamCpExecSubprocess):
    def __init__(self, args={}):
        SamCpExecSubprocess.__init__(self, args)


############################################################################
# New enstore encp class.

class SamEncpEnstore(SamCpEnstore):
    def __init__(self, args={}):
        SamCpEnstore.__init__(self, args)
        self.command = "${SAM_ENCP_DIR}/bin/sam_encp"

    ########################################################################
    # pre-processing 
    def prepareFileTransfer(self):
        SamCpEnstore.prepareFileTransfer(self)

    ########################################################################
    # actual file transfer
    def performFileTransfer(self):
        import SamEncpWrapper
        src  = os.path.expandvars("%s" % self.srcFile)
        dest = os.path.expandvars("%s" % self.destFile)
        args = {}
        args['srcFileString'] = src
        args['destFileString'] = dest
        args['debug'] = self.debug
        args['verbose'] = self.verbose
        encpWrapper = SamEncpWrapper.SamEncpWrapper(args)
        self.fullCommand = encpWrapper.getExecCommand()
        self.dbgprint("%s performFileTransfer: Will execute: %s" % (
            self.className, self.fullCommand))
        
        SamCpEnstore.performFileTransfer(self)
        
    ########################################################################
    # post-processing 
    def postProcessFileTransfer(self):
        if ( not self.dest.isEnstore() ):
            self.dbgprint("%s postProcess: destination is not enstore" % (
                self.className, srcFile))
            return


        # Get info about a file.
        srcFile = self.src.basename()
        targetFile = self.dest.path() + "/" + srcFile
        self.dbgprint("%s postProcessFileTransfer: srcFile = %s" % (
          self.className, srcFile))
        self.dbgprint("%s postProcessFileTransfer: targetFile = %s" % (
          self.className, targetFile))
        fileInfoHandler = SamCpEnstoreFileInfoHandler.SamCpEnstoreFileInfoHandler(targetFile)
        crcValue = fileInfoHandler.getCRCValue()
        volumeLabel = fileInfoHandler.getVolumeLabel()
        volumeOffset = fileInfoHandler.getVolumeOffset()
        outputDict = {
          SamCpOutputHandler.ENSTORE_CRC_VALUE_KEY_ : crcValue,
          SamCpOutputHandler.ENSTORE_VOLUME_LABEL_KEY_ : volumeLabel,
          SamCpOutputHandler.ENSTORE_VOLUME_OFFSET_KEY_ : volumeOffset,
          }
        self.outputHandler.add(outputDict)
        self.dbgprint("%s postProcessFileTransfer: exitStatus: %s" % (
          self.className, self.exitStatus))
        SamCpExecSubprocess.postProcessFileTransfer(self)

        
############################################################################ 
# DCAP for Enstore (dcache stuff, new package):

class SamDccpEnstore(SamCpEnstore):

    class AcquireLockTimeout(SamCpExceptions.SamCpException):
        def __init__(self, msg):
            self.args = msg

    """Args behave like a local copy"""
    def __init__(self, args={}):
        SamCpEnstore.__init__(self, args)
        self.command = "${SAM_DCCP_DIR}/bin/sam_dccp.sh"
        self.postProcessCommand = "${SAM_DCCP_DIR}/bin/sam_dc_stage.sh"
        self.dccpLimitEnvVar = "SAM_DCCP_LIMIT"
        self.dccpLimitTimeOutEnvVar = "SAM_DCCP_LIMIT_TIMEOUT"
        self.minSleepTime = 60
        self.maxSleepTime = 120
        self.lockFileDir = "/tmp"
        self.lockFilePrefix = "sam_dccp_lock"
        self.lockTimeOut = int(os.environ.get(self.dccpLimitTimeOutEnvVar, 3600)) 
        self.lockFile = None
        self.nProcMax = int(os.environ.get(self.dccpLimitEnvVar, 0)) 

    def cleanup(self, foundLockFiles):

        t = time.time()
        for file in foundLockFiles:
            lockFile = "%s/%s" % (self.lockFileDir, file)
            msg = "%s throttle: checking lock timeout for file %s" % (self.className, file)
            self.dbgprint(msg)
            try:
                fileMTime = os.stat(lockFile)[stat.ST_MTIME]
                if t - fileMTime > self.lockTimeOut:
                    msg = "%s throttle: lock timeout expired for file %s" % (self.className, file)
                    self.dbgprint(msg)
                    try:
                        os.remove(lockFile)
                    except:
                        msg = "Cannot remove expired lock file %s: %s" % (file, sys.exc_info()[1])
                        self.dbgprint("%s throttle: %s" % (self.className, msg))                                
            except:
                msg = "Cannot stat lock file %s: %s" % (file, sys.exc_info()[1])
                self.dbgprint("%s throttle: %s" % (self.className, msg))


    def throttle(self):
        """ Throttle number of simultaneous dccp processes.
        Hack for the online troubles. """
        try:
            self.dbgprint("%s throttle: method called" % (
                self.className))
            user = getUnixUsername()
            pid = os.getpid()
            self.dbgprint("%s throttle: max. number of dccp processes: %s" % (
                self.className, self.nProcMax))
            if self.nProcMax > 0:
                # Limit the number of dccp's."
                # The idea is that there is n dccp slots, and each
                # slot has a lock file associated with it.
                # Before the transfer begins, sam_cp will acquire 
                # a lock on one of those files, and remove it
                # after the transfer ends. 
                # There is also a lock timeout concept,
                # meaning that after a certain period of time transfer
                # proceeds regardless of the lock file ownership. 
                startTime = time.time()
                while 1:
                    acquiredLock = 0
                    lockDirFiles = os.listdir(self.lockFileDir)
                    foundLockFiles = []
                    for i in range (0, self.nProcMax):
                        lockFilePattern = "%s.%s" % (self.lockFilePrefix, i)
                        msg = "%s throttle: looking for lock file pattern %s" % (self.className, lockFilePattern)
                        self.dbgprint(msg)
                        foundLockFile = 0
                        for file in lockDirFiles:
                            if re.match(lockFilePattern, file):
                                foundLockFile = 1
                                foundLockFiles.append(file)
                                break
                        if not foundLockFile:
                            acquiredLock = 1
                            break
                    if acquiredLock:
                        break
                    else:
                        endTime = time.time()
                        deltaT = endTime - startTime
                        if deltaT > self.lockTimeOut:
                            # Raise exception and fail.
                            srcFile = self.src.basename()
                            msg = "Cannot acquire dccp lock for file %s" % (srcFile)
                            self.dbgprint("%s throttle: %s" % (self.className, msg))

                            ## No more waiting for lock. Cleanup stale lock files and fail.
                            self.cleanup(foundLockFiles)

                            raise SamDccpEnstore.AcquireLockTimeout(msg)


                        else:
                            # Timeout has not passed, wait some more.
                            sleepTime = random.uniform(self.minSleepTime, self.maxSleepTime)
                            msg = "%s throttle: could not acquire lock file, will wait %s seconds" % (self.className, sleepTime)
                            self.dbgprint(msg)
                            time.sleep(sleepTime)

                # Done waiting, get lock.
                self.lockFile = "%s/%s.%s.%s" % (self.lockFileDir, lockFilePattern, pid, user)
                open(self.lockFile, "w").close()
                msg = "%s throttle: acquired lock file %s" % (self.className, self.lockFile)
                self.dbgprint(msg)
            else:
                self.dbgprint("%s throttle: no limit on the number of dccp processes" % (
                    self.className))
        except:
            # we failed, for any reason, we gotta pass it on up.
            msg = "Throttling failed: %s" % (sys.exc_info()[1])
            self.dbgprint("%s throttle: %s" % (
                    self.className, msg))
            raise SamCpExceptions.SamCpFailure(msg)
        
    ########################################################################
    # pre-processing 
    def prepareFileTransfer(self):
        SamCpEnstore.prepareFileTransfer(self)
        self.throttle()
        
    ########################################################################
    # post-processing 
    def postProcessFileTransfer(self):
        if self.nProcMax > 0:
            # Throttling was done, cleanup here.
            try:
                os.remove(self.lockFile)
            except:
                msg = "Cannot remove expired lock file %s: %s" % (file, sys.exc_info()[1])
                self.dbgprint("%s throttle: %s" % (self.className, msg))

        if ( not self.dest.isEnstore() ):
            self.dbgprint("%s postProcess: destination is not enstore" % (
                self.className, srcFile))
            return


        # Get info about a file.
        srcFile = self.src.basename()
        targetFile = self.dest.path() + "/" + srcFile
        self.dbgprint("%s postProcessFileTransfer: srcFile = %s" % (
          self.className, srcFile))
        self.dbgprint("%s postProcessFileTransfer: targetFile = %s" % (
          self.className, targetFile))

        fullCmd = "%s %s" % (self.postProcessCommand, targetFile)
        self.dbgprint("%s exec> %s" % (self.className, fullCmd))
        (stdOut, stdErr, exitStatus) = ExecSubprocess(fullCmd)
        self.dbgprint("%s postProcessFileTransfer: exitStatus: %s" % (
          self.className, exitStatus))
        self.dbgprint("%s postProcessFileTransfer: stdOut: %s" % (
          self.className, stdOut))
        self.dbgprint("%s postProcessFileTransfer: stdErr: %s" % (
          self.className, stdErr))
        outputDict = {
          SamCpOutputHandler.POST_PROCESS_EXIT_STATUS_KEY_ : exitStatus,
          SamCpOutputHandler.POST_PROCESS_STD_ERR_KEY_ : stdErr,
          SamCpOutputHandler.POST_PROCESS_STD_OUT_KEY_ : stdOut,
          }
        self.outputHandler.add(outputDict)

        fileInfoHandler = SamCpEnstoreFileInfoHandler.SamCpEnstoreFileInfoHandler(targetFile)

        # Try to get crc value. fileinfo is broken at the moment, so exception
        # will be caught.
        crcValue = SamCpOutputHandler.UNKNOWN_DCACHE_CRC_VALUE_
        try:
          crcValue = fileInfoHandler.getCRCValue()
        except SamCpEnstoreFileInfoHandler.SamCpEnstoreFileInfoHandlerException, ex:
          self.dbgprint("%s postProcessFileTransfer: cannot determine crc value: %s" % (
            self.className, "%s" % (ex)))

        outputDict = {
          SamCpOutputHandler.ENSTORE_CRC_VALUE_KEY_ : crcValue,
          SamCpOutputHandler.ENSTORE_VOLUME_LABEL_KEY_ : SamCpOutputHandler.UNKNOWN_DCACHE_VOLUME_LABEL_,
          SamCpOutputHandler.ENSTORE_VOLUME_OFFSET_KEY_ : SamCpOutputHandler.UNKNOWN_DCACHE_VOLUME_OFFSET_,
          }
        self.outputHandler.add(outputDict)

        SamCpEnstore.postProcessFileTransfer(self)
        if ( exitStatus ):
            msg = "Post process failed: %s " % (stdErr)
            raise SamCpExceptions.SamCpPostProcessFailure(msg)

        if not self.exitStatus:
            self.exitStatus = exitStatus
        else:
            self.exitStatus = self.exitStatus | exitStatus
        self.dbgprint("%s postProcessFileTransfer: exitStatus: %s" % (
          self.className, self.exitStatus))
        SamCpExecSubprocess.postProcessFileTransfer(self)

class SamCmsDcacheGridftp(SamCpExecSubprocess):
    def __init__(self, args={}):
        SamCpExecSubprocess.__init__(self, args)
        self.command = "${SAM_DCACHE_CP_DIR}/bin/sam_dcache_gridftp"
        self.ownerName = None
        self.datasetName = None

    def prepareFileTransfer(self):
	# This will work only for reads from dcache.
 	# Need to save the original source file.
        self.origSrcFile = self.filterFileName(self.src.parsedFileName())
        self.srcFile = self.getSourceFileName(self.origSrcFile)
        self.destFile = self.dest.parsedFileName()
 
        self.dbgprint("prepareFileTransfer: origSrcFile = %s" % self.origSrcFile)
        self.dbgprint("prepareFileTransfer: srcFile = %s" % self.srcFile)
        self.dbgprint("prepareFileTransfer: destFile = %s" % self.destFile)

        self.args = []
        self.dbgprint("prepareFileTransfer: args = %s" % str(self.args))

    def postProcessFileTransfer(self):
	# This will have to move the copied file intt the originally requested
	# file in order not to confuse the station.
	from os.path import basename, dirname
	destDir = self.destFile
	copiedFileName = basename(self.srcFile)
	copiedDestFile = destDir + "/" + copiedFileName
	origFileName = basename(self.origSrcFile)
	origDestFile = destDir + "/" + origFileName
        self.dbgprint("postProcessFileTransfer: copiedDestFile = %s" % copiedDestFile)
        self.dbgprint("postProcessFileTransfer: origDestFile = %s" % origDestFile)
	# Move the copied file into the original one.
        self.dbgprint("postProcessFileTransfer: moving %s into %s" % (copiedDestFile, origDestFile))
	os.rename(copiedDestFile, origDestFile)
        # Put in a directory tree and links based on filename back to the
        # SAM saved files named by GUID. Go one level up the easy way for now.
        lpath = destDir+"/../"+self.ownerName+"/"+self.datasetName
        try:
            os.makedirs(lpath)
        except OSError:
            pass
        try:
            os.symlink(origDestFile,lpath+"/"+copiedFileName)
        except:
            self.dbgprint("symlink failure:"+origDestFile+"->"+lpath+"/"+copiedFileName)
            pass
    def filterFileName(self, fileName):
        # Look for stuff like (tape,offset) and get rid of it.
	import re
        self.dbgprint("filterFileName: original fileName = %s" % fileName)
	newFileName = re.compile('\(.*\)').sub('', fileName)
        self.dbgprint("filterFileName: filtered fileName = %s" % newFileName)
        return newFileName

    def getSourceFileName(self, fullFileName):
        # Assume only sam_common_pylib is setup, so we have
 	# to do things the hard way: get DB server proxy,
	# construct arguments, get metadata, and extract LogicalName.
	from os.path import basename, dirname
	from SamCorba import DbServerProxy
	from SamStruct import NameOrId
	from SamStruct import CorbaCaseInsensitiveDictionary
	
        # This is the wrapper struct for passing to the db server.
	# Using wrapper struct is a bit easier than actual corba structs.
        dirName = dirname(fullFileName)
        fileName = basename(fullFileName)
        self.dbgprint("getSourceFileName: fullFileName = %s" % fullFileName)
        self.dbgprint("getSourceFileName: dirName = %s" % dirName)
        self.dbgprint("getSourceFileName: fileName = %s" % fileName)

	file = NameOrId.NameOrId(fileName)

	# This is the DB Server proxy.
	dbs = DbServerProxy.DbDataFileProxy()

	# Get metadata. If any exception occurs, it will not be handled
	# properly at this time and the transfer will fail.
	# corbaRepr() gives us appropriate corba struct for the call.
	fileMD = dbs.getServant().getMetadata(file.corbaRepr())

	# Extract the LogicalName. We need dictionary wrapper to make
	# things a bit easier. Otherwise we would have to unpack the raw
	# corba struct.
	attrDict = CorbaCaseInsensitiveDictionary.CorbaCaseInsensitiveDictionary(fileMD.additionalAttributes)
	logicalName = attrDict.data.get('params').data.get('CMSGlobalCategory').data.get('LogicalName')
        self.dbgprint("getSourceFileName: logicalName = %s" % logicalName)
        ## added to help with setting up a reasonable structure for links
        self.ownerName = attrDict.data.get('params').data.get('CMSGlobalCategory').data.get('Owner')
        self.datasetName = attrDict.data.get('params').data.get('CMSGlobalCategory').data.get('Dataset')

	# Done, return the new full file name.
	result = dirName + "/" + logicalName	
        self.dbgprint("getSourceFileName: sourceFileName = %s" % result)
	return result
	


################################################################################
    
if __name__ == "__main__":

    #print("testing base class:")
    #baseClass = SamCp()
    #baseClass.copy("param1", "param2", ["-alx"])

    print("testing dccp class:")
    dccpClass = SamDccp()
    dccpClass.copy("/pnfs/sam/dzero/demo/sam_test_file.ev", "/scratch/7/veseli")

