#!/usr/bin/env sampy
################################################################################
# Project   : SAM
# Package : sam_cp
# $Id: sam_cp.py,v
# 1.29 2001/06/04 21:39:09 lauri Exp $
#
# This work is part of a development project, called SAM, which consists of a
# number of coordinated packages each named sam_xxxx .
#
# Notice of authorship, copyright status,  and terms and conditions, should
# the software eventually become available for use outside Fermilab, can be
# found in the README and LICENCE files in the top level directory of the main
# sam package.
################################################################################

import sys
import os
import os.path
import imp
import time
import string
import traceback

from SamCpImportManager import getSamCpImportManager
from SamCpFileParser import SamCpFileParser
import SamCpExceptions
from SamStruct.SamBoolean import SamBoolean

GOOD_RETURN_STATUS_ = 1
BAD_RETURN_STATUS_ = 0

TRUE_ = 1
FALSE_ = 0

# the names of the environmental variables we will examine
SAM_CP_CONFIG_FILE_                     = "SAM_CP_CONFIG_FILE"
SAM_CP_USER_SCRIPT_                     = "SAM_CP_USER_SCRIPT"
SAM_CP_ARBITRATOR_MODULE_FILE_          = "SAM_CP_ARBITRATOR_MODULE_FILE"
SAM_CP_ARBITRATOR_CLASS_NAME_           = "SAM_CP_ARBITRATOR_CLASS_NAME"
SAM_CP_ARBITRATOR_CONFIG_FILE_          = "SAM_CP_ARBITRATOR_CONFIG_FILE"
SAM_CP_CONFIG_MGR_MODULE_FILE_          = "SAM_CP_CONFIG_MGR_MODULE_FILE"
SAM_CP_CONFIG_MGR_CLASS_NAME_           = "SAM_CP_CONFIG_MGR_CLASS_NAME"

# the default values we will use
DEFAULTS_ = { SAM_CP_ARBITRATOR_MODULE_FILE_ : "${SAM_CP_DIR}/lib/SamCpProtocolArbitrator.py",
              SAM_CP_ARBITRATOR_CLASS_NAME_  : "SamCpDefaultProtocolArbitrator",
              SAM_CP_ARBITRATOR_CONFIG_FILE_ : "",
              SAM_CP_CONFIG_MGR_MODULE_FILE_ : "${SAM_CP_DIR}/lib/SamCpConfigManager.py",
              SAM_CP_CONFIG_MGR_CLASS_NAME_  : "SamCpDefaultConfigManager",
              }

# the keywords for the argDict
DEBUG_                   = 'd'
VERBOSE_                 = 'v'
SRC_                     = 'src'
DEST_                    = 'dest'
USE_                     = 'use'
CONFIG_FILE_             = 'configFile'
CONFIG_MGR_MODULE_       = 'configMgrModule'
CONFIG_MGR_CLASS_        = 'configMgrClass'
ARBITRATION_MODULE_      = 'arbitrationModule'
ARBITRATION_CLASS_       = 'arbitrationClass'
ARBITRATION_CONFIG_FILE_ = 'arbitrationConfigFile'


########################################################################################
#==================================================

from SamCommand.CommandInterface import CommandInterface

class SamCpWrapper(CommandInterface):
    def __init__(self):
        CommandInterface.__init__(
            self,
            commandParameters={
            'verb' : [ '', ],
            'allowedOptions' : ['sam-cp-user-script=',
                                'use=',
                                'config=',
                                'config-mgr-module=', 'config-mgr-class=',
                                'arbitrator-module=', 'arbitrator-class=',
                                'arbitrator-config-file='],
            'allowedArgCount' : 2,
            'allowedFlags' : ['d', 'v',],
            'description' :
            { 'args' : 'srcFile destFile',
              'd' : 'debugging mode',
              'v' : 'verbose mode',
              'sam-cp-user-script': 'entirely bypass sam_cp, call your script instead',
              'use' : 'force sam_cp to use this protocol (sam_bbftp, sam_kerberos_rcp, etc.)',
              'config' : 'overrides env setting of $%s' % SAM_CP_USER_SCRIPT_,
              'configMgrModule' : 'overrides env setting of $%s' % SAM_CP_CONFIG_MGR_MODULE_FILE_,
              'configMgrClass' : 'overrides env setting of $%s' % SAM_CP_CONFIG_MGR_CLASS_NAME_,
              'arbitratorModule' : 'overrides env setting of $%s' % SAM_CP_ARBITRATOR_MODULE_FILE_,
              'arbitratorClass' : 'overrides env setting of $%s' % SAM_CP_ARBITRATOR_CLASS_NAME_,
              'arbitratorConfigFile' : 'overrides env setting of $%s' % SAM_CP_ARBITRATOR_CONFIG_FILE_,
              },
            'helpText' : """
    sam_cp is a utility used by SAM stations and fss's for
    transferring files between stations.  By default it uses
    a very simple (and unscaleable) SAM_CP_CONFIG_FILE
    configuration based on domain-name and hard-coded
    capabilities defined for each relevent domain.

    Normal usage:
         setup sam_cp
         sam_cp srcFile destFile

    For those who are adventurous, the following is a
    brief summary of the hooks available for you to
    augment or bypass sam_cp:

       $SAM_CP_USER_SCRIPT:
           Entirely by-pass sam_cp and pass control to
           the specified script.
           
       $SAM_CP_CONFIG_MGR_MODULE_FILE:
           full path to the module containing a
           configuration manager class, wherein is the
           logic to parse $SAM_CP_CONFIG_FILE

       $SAM_CP_CONFIG_MGR_CLASS_NAME:
           name of the class which performs the parsing of the
           $SAM_CP_CONFIG_FILE.  It must implement
           the following routines, with the specified
           return types:
             stringList = cMgr.getKnownSamCpProtocolNames()
             stringList = cMgr.getLocallyAvailableSamCpProtocolNames() 
             string     = cMgr.getPreferredSamCpProtocolName()
             samCpClass = cMgr.getProtocolClass(protocolName)

       $SAM_CP_CONFIG_FILE:
           if you choose to write your own configuration manager,
           you are free to choose an alternate name (and format)
           for the $SAM_CP_CONFIG_FILE.

       $SAM_CP_ARBITRATOR_MODULE_FILE:
           full path to the python module containing
           a protocol arbitrator class that contains
           the necessary logic to select the most
           appropriate protocol

       $SAM_CP_ARBITRATOR_CLASS_NAME:
           name of the class which contains the algorithm
           for selecting the most appropriate file
           transfer protocol.  It must implement
           the following routine:
              samCpProtocolName = pMgr.selectProtocol()

       $SAM_CP_ARBITRATOR_CONFIG_FILE:
           full path to a file containing any configurations
           (beyond those in the basic $SAM_CP_CONFIG_FILE) that
           are needed by the arbitrator class.  (That is, this
           is a hook for people who are writing their own
           protocol arbitrator).

    For further information, RTFC (aka Read The Fine Code,
    found in the ${SAM_CP_DIR}/lib directory).

    Example of completely bypassing sam_cp:
        setup sam_cp
        export SAM_CP_USER_SCRIPT=${SAM_CP_DIR}/examples/sam_cp_user_script_example.sh

    Example of an alternate protocol arbitrator:
        setup sam_cp
        export SAM_CP_ARBITRATOR_MODULE_FILE=${SAM_CP_DIR}/examples/sam_cp_protocol_arbitrator.py
        export SAM_CP_ARBITRATOR_CLASS_NAME="my_protocol_arbitrator"
        sam_cp srcFile destFile

    A better example of an alternate protocol arbitrator,
    one which actually does something:
        setup sam_cp
        export SAM_CP_ARBITRATOR_MODULE_FILE="${SAM_CP_DIR}/examples/SamCpMetadataProtocolArbitrator.py"
        export SAM_CP_ARBITRATOR_CLASS_NAME="SamCpMetadataProtocolArbitrator"
        export SAM_CP_ARBITRATOR_CONFIG_FILE="${SAM_CP_DIR}/examples/sam_cp_metadata_protocol_arbitrator_config.py"
        sam_cp srcFile destFile
        
    Example of an alternate config manager:
        setup sam_cp
        export SAM_CP_CONFIG_MGR_MODULE_FILE=${SAM_CP_DIR}/examples/sam__cp_config_manager.py
        export SAM_CP_CONFIG_MGR_CLASS_NAME="my_config_manager"
        sam_cp srcFile destFile
    """,})

    def implementation(self, argDict, argList):
        # parse the incoming argdict
        self.parseCommandArgs(argDict, argList)
        self.importManager = getSamCpImportManager()
        self.configMgr = None
        self.arbitrator = None
        return self.execute()

    #########################################################################
    def parseCommandArgs(self, argDict, argList):
        self.debug = SamBoolean(argDict.get('d', 0))
        self.verbose = SamBoolean(argDict.get('v', 0))
        self.useThisProtocol = argDict.get(USE_, None)
        self.dbgprint("INCOMING SamCpRequest: argDict = %s" % argDict)

        self.configFile      = argDict.get(CONFIG_FILE_,        os.environ.get(SAM_CP_CONFIG_FILE_))
        self.configMgrModule = argDict.get(CONFIG_MGR_MODULE_,  os.environ.get(SAM_CP_CONFIG_MGR_MODULE_FILE_))
        self.configMgrClass  = argDict.get(CONFIG_MGR_CLASS_,   os.environ.get(SAM_CP_CONFIG_MGR_CLASS_NAME_))
        self.dbgprint("configFile = %s" % self.configFile)
        self.dbgprint("configMgrModule = %s" % self.configMgrModule)
        self.dbgprint("configMgrClass = %s" % self.configMgrClass)
        
        self.arbitrationModule     = argDict.get(ARBITRATION_MODULE_, os.environ.get(SAM_CP_ARBITRATOR_MODULE_FILE_))
        self.arbitrationClass      = argDict.get(ARBITRATION_CLASS_,  os.environ.get(SAM_CP_ARBITRATOR_CLASS_NAME_))
        self.arbitrationConfigFile = argDict.get(ARBITRATION_CONFIG_FILE_, os.environ.get(SAM_CP_ARBITRATOR_CONFIG_FILE_))
        self.dbgprint("arbitrationModule = %s" % self.arbitrationModule)
        self.dbgprint("arbitrationClass = %s" % self.arbitrationClass)
        self.dbgprint("arbitrationConfigFile = %s" % self.arbitrationConfigFile)

        self.srcFileString  = argList[0]
        self.destFileString = argList[1]
        self.dbgprint("src = %s" % self.srcFileString)
        self.dbgprint("dest = %s" % self.destFileString)

        self.curLoc  = SamCpFileParser()
        self.srcLoc  = SamCpFileParser(self.srcFileString)
        self.destLoc = SamCpFileParser(self.destFileString)
        self.dbgprint("curLoc = %s" % self.curLoc)
        self.dbgprint("srcLoc = %s" % self.srcLoc)
        self.dbgprint("destLoc = %s" % self.destLoc)

    #########################################################################

    def getDebug(self):                  return self.debug
    def getVerbose(self):                return self.verbose
    def getUseThisProtocol(self):        return self.useThisProtocol
    def getConfigFile(self):             return self.configFile
    def getSrcFileString(self):          return self.srcFileString
    def getDestFileString(self):         return self.destFileString
    def getDbgprint(self):               return self.dbgprint
    def getVsay(self):                   return self.vsay
    def getCurLoc(self):                 return self.curLoc
    def getSrcLoc(self):                 return self.srcLoc
    def getDestLoc(self):                return self.destLoc
    def getConfigMgrModule(self):        return self.configMgrModule
    def getConfigMgrClass(self):         return self.configMgrClass
    def getConfigMgr(self):              return self.configMgr
    def getArbitrationModule(self):      return self.arbitrationModule
    def getArbitrationClass(self):       return self.arbitrationClass
    def getArbitrationConfigFile(self):  return self.arbitrationConfigFile
    def getArbitrator(self):             return self.arbitrator

    #########################################################################
    def execute(self):
        self.vsay("sam_cp %s %s" % (self.srcFileString, self.destFileString))

        # determine the configuration manager to use
        try:
            self.configMgr = self.importConfigManager()
            self.dbgprint("selected configManager = %s" % str(self.configMgr))
        except:
            raise SamCpExceptions.SamCpException(
                "Cannot import ConfigManager class '%s' from module '%s':\n%s" %
                (self.configMgrClass, self.configMgrModule, sys.exc_info()[1]))

        # determine the arbiter, that is, the algorithm that will be used
        #  to select the appropriate sam_cp protocol
        try:
            self.arbitrator = self.importArbitrator()
            self.dbgprint("selected arbitrator = %s" % str(self.arbitrator))
        except:
            raise SamCpExceptions.SamCpException(
                "Cannot import ProtocolArbitrator class '%s' from module '%s':\n%s" %
                (self.arbitrationClass, self.arbitrationModule, sys.exc_info()[1]))

        # let the arbitrator select the appropriate protocol name, based on our configuration
        try:
            protocolName = self.arbitrator.selectProtocol()
            self.dbgprint("arbitrator selected protocol name = %s (type %s)" % (protocolName, type(protocolName)))
        except:
            raise SamCpExceptions.SamCpException(
                "No protocolName selected:\n%s" % sys.exc_info()[1])

        try:
            # now find the class corresponding to this protocol name
            protocol = self.configMgr.getProtocolClass(protocolName)

            # create an instance of the protocol selected by the arbitrator
            copier = self.createCopier(protocol)
            self.dbgprint("created copier = %s (type %s)" % (copier, type(copier)))

            sts = copier.copy()
            return sts
        except (SamCpExceptions.SamCpException), ex:
            raise
        except:
            print "%s" % sys.exc_info()[1]
            raise SamCpExceptions.SamCpFailure("%s" % sys.exc_info()[1])

    #########################################################################
    def importArbitrator(self):
        # import the class
        arbitratorClass = self.importManager.importObjectFromFile(self.arbitrationClass, self.arbitrationModule)

        # create an arbitrator instance
        arbitrator = arbitratorClass(self)

        # return the arbitrator
        return arbitrator

    #########################################################################
    def importConfigManager(self):
        # import the class
        configMgrClass = self.importManager.importObjectFromFile(self.configMgrClass, self.configMgrModule)

        # create a configMgr instance
        configMgr = configMgrClass(self)

        # return the configMgr
        return configMgr

    #########################################################################
    def createCopier(self, specificSamCpProtocolClass):
        argDict = { SRC_ : self.srcLoc,
                    DEST_ : self.destLoc,
                    DEBUG_ : self.debug,
                    VERBOSE_ : self.verbose }
        copierInstance = specificSamCpProtocolClass(argDict)
        return copierInstance
    #########################################################################
    def dbgprint(self, astring):
        if ( self.debug ):
	    now = time.strftime( '%H:%M:%S', time.localtime(time.time()) )
            print "<%s> %s" % (now,astring)

    #########################################################################
    def vsay(self, astring):
        if ( self.verbose ):
            print "%s" % astring
#===============================================================================
#===============================================================================
def selftest():
    srcFile = "datalogger:d0lb:/jerry/sam_demo_file.ev"
    destFile = "enstore:/this_is_my_dir/sam_demo_file.ev"
    configDir = "/scratch/d0test/veseli/sam_devel/online"
    configModule = "sam_cp_online_config"
    for protocol in ('local_cp',
                     'sam_bbftp',
                     'sam_kerberos_rcp',
                     'enstore',
                     'encp',
                     'dccp',
                     'rcp'):
        try:
            print("Testing protocol %s..." % protocol)
            status = SamCpWrapper()(use=protocol,
                                    d=1,
                                    v=1,
                                    samCpConfigFile=os.environ['SAM_CP_CONFIG_FILE'],
                                    args=['/this/is/a/test/src/file',
                                          '/this/is/a/test/dest/file'])
            print("Status = %s" % status)
        except:
            print("Exception: %s %s" % (sys.exc_info()[0], sys.exc_info()[1]))
            #traceback.print_tb(sys.exc_info()[2])
            print(80*"=")

#===============================================================================
#===============================================================================

def main(args):
    return SamCpWrapper().dispatch(args)

if __name__ == "__main__":
    if len(sys.argv) == 2 and sys.argv[1] == 'test':
        sys.exit(selftest())
    sys.exit(main(sys.argv[1:]))
